/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/GA/src/CrossoverBlendESVecOp.cpp
 *  \brief  Source code of class GA::CrossoverBlendFltVec.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.2 $
 *  $Date: 2004/04/25 23:12:19 $
 */

#include "beagle/GA.hpp"

#include <algorithm>
#include <string>

using namespace Beagle;


/*!
 *  \brief Construct a ES blend crossover operator.
 *  \param inMatingPbName Mating probability parameter name.
 *  \param inName Name of the operator.
 */
Beagle::GA::CrossoverBlendESVecOp::CrossoverBlendESVecOp(std::string inMatingPbName,
                                                         std::string inName) :
  CrossoverOp(inMatingPbName, inName)
{ }


/*!
 *  \brief Initialize the ES blend crossover operator.
 *  \param ioSystem System of the evolution.
 */
void Beagle::GA::CrossoverBlendESVecOp::initialize(Beagle::System& ioSystem)
{
  CrossoverOp::initialize(ioSystem);

  if(ioSystem.getRegister().isRegistered(mMatingProbaName)) {
     ioSystem.getRegister().deleteEntry(mMatingProbaName);
  }

  if(ioSystem.getRegister().isRegistered(mMatingProbaName)) {
    mMatingProba = castHandleT<Float>(ioSystem.getRegister()[mMatingProbaName]);
  } else {
    mMatingProba = new Float(float(0.3));
    Register::Description lDescription(
      "Individual blend crossover prob.",
      "Float",
      "0.3",
      "ES blend crossover probability of a single individual."
    );
    ioSystem.getRegister().addEntry(mMatingProbaName, mMatingProba, lDescription);
  }
}


/*!
 *  \brief Mate two ES individuals for blend crossover.
 *  \param ioIndiv1   First individual to mate.
 *  \param ioContext1 Evolutionary context of the first individual.
 *  \param ioIndiv2   Second individual to mate.
 *  \param ioContext2 Evolutionary context of the second individual.
 *  \return True if the individuals are effectively mated, false if not.
 */
bool Beagle::GA::CrossoverBlendESVecOp::mate(Beagle::Individual& ioIndiv1,
                                             Beagle::Context&    ioContext1,
                                             Beagle::Individual& ioIndiv2,
                                             Beagle::Context&    ioContext2)
{
  unsigned int lNbGenotypes = minOf<unsigned int>(ioIndiv1.size(), ioIndiv2.size());
  if(lNbGenotypes == 0) return false;

  Beagle_LogDebugM(
    ioContext1.getSystem().getLogger(),
    "crossover", "Beagle::GA::CrossoverBlendESVecOp",
    std::string("The first individual mated is (before ES blend crossover): ")+
    ioIndiv1.serialize()
  );
  Beagle_LogDebugM(
    ioContext1.getSystem().getLogger(),
    "crossover", "Beagle::GA::CrossoverBlendESVecOp",
    std::string("The second individual mated is (before ES blend crossover): ")+
    ioIndiv2.serialize()
  );

  for(unsigned int i=0; i<lNbGenotypes; ++i) {
    GA::ESVector::Handle lESVector1 = castHandleT<ESVector>(ioIndiv1[i]);
    GA::ESVector::Handle lESVector2 = castHandleT<ESVector>(ioIndiv2[i]);
    unsigned int lSize = minOf<unsigned int>(lESVector1->size(), lESVector2->size());
    double lAlpha = ioContext1.getSystem().getRandomizer().rollUniform();
    for(unsigned int j=0; j<lSize; ++j) {
      double lDiffValue = lAlpha * ((*lESVector2)[j].mValue - (*lESVector1)[j].mValue);
      (*lESVector1)[j].mValue += lDiffValue;
      (*lESVector2)[j].mValue -= lDiffValue;
      double lDiffStrategy = lAlpha * ((*lESVector2)[j].mStrategy - (*lESVector1)[j].mStrategy);
      (*lESVector1)[j].mStrategy += lDiffStrategy;
      (*lESVector2)[j].mStrategy -= lDiffStrategy;
    }
  }

  Beagle_LogDebugM(
    ioContext1.getSystem().getLogger(),
    "crossover", "Beagle::GA::CrossoverBlendEsVecOp",
    std::string("The first individual mated is (after ES blend crossover): ")+
    ioIndiv1.serialize()
  );
  Beagle_LogDebugM(
    ioContext1.getSystem().getLogger(),
    "crossover", "Beagle::GA::CrossoverBlendEsVecOp",
    std::string("The second individual mated is (after ES blend crossover): ")+
    ioIndiv2.serialize()
  );

  return true;
}


