/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/GA/src/InitESVecOp.cpp
 *  \brief  Source code of class GA::InitESVecOp.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.5 $
 *  $Date: 2004/07/04 16:58:22 $
 */

#include "beagle/GA.hpp"

#include <sstream>

using namespace Beagle;


/*!
 *  \brief Construct an ES vectors initialization operator.
 *  \param inESVectorSize Size of the ES vectors initialized.
 *  \param inReproProbaName Reproduction probability parameter name used in register. 
 *  \param inName Name of the operator.
 */
GA::InitESVecOp::InitESVecOp(IntegerVector inESVectorSize,
                             std::string inReproProbaName,
                             std::string inName) :
  InitializationOp(inReproProbaName, inName),
  mESVectorSize(new IntegerVector(inESVectorSize))
{ }


/*!
 *  \brief Initialize the ES vectors initialization operator.
 *  \param ioSystem System of the evolution.
 */
void GA::InitESVecOp::initialize(System& ioSystem)
{
  InitializationOp::initialize(ioSystem);

  if(ioSystem.getRegister().isRegistered("es.init.vectorsize")) {
    mESVectorSize = castHandleT<IntegerVector>(ioSystem.getRegister()["es.init.vectorsize"]);
  } else {
    std::string lDefaultValue;
    for(unsigned int i=0; i<mESVectorSize->size(); i++) {
      lDefaultValue += uint2str((*mESVectorSize)[i]);
      if(i != (mESVectorSize->size()-1)) lDefaultValue += "/";
    }
    std::string lLongDescrip("ES vectors size for each individual. The size of the ");
    lLongDescrip += "IntegerVector is the number of vectors composing the ES individuals, ";
    lLongDescrip += "while each value of the vector is the number of ES pairs composing each ";
    lLongDescrip += "vector. IntegerVector format is S1/S2/.../Sn, where Si is ith value.";
    Register::Description lDescription(
      "Initial ES vectors size",
      "IntegerVector",
      lDefaultValue,
      lLongDescrip
    );
    ioSystem.getRegister().addEntry("es.init.vectorsize", mESVectorSize, lDescription);
  }

  if(ioSystem.getRegister().isRegistered("es.init.strategy")) {
    mInitStrategyValue = castHandleT<Float>(ioSystem.getRegister()["es.init.strategy"]);
  } else {
    mInitStrategyValue = new Float(1.0);
    Register::Description lDescription(
      "Initial strategy parameter value",
      "Float",
      "1.0",
      "Initial strategy parameter value used to initialize ES vectors."
    );
    ioSystem.getRegister().addEntry("es.init.strategy", mInitStrategyValue, lDescription);
  }
}


/*!
 *  \brief Initialize ES individual with initialial strategy parameter.
 *  \param outIndividual Individual to initialize.
 *  \param ioContext Evolution context.
 *
 *  Size of ES individual is given by integer vector "es.init.vectorsize". Initial strategy
 *  parameter value of ES pair is given by parameter "es.init.strategy". Initial value of ES pair
 *  is given using random numbers following a Gaussian distribution of zero mean and standard
 *  deviation equal to initial strategy parameter value.
 */
void GA::InitESVecOp::initIndividual(Beagle::Individual& outIndividual, Context& ioContext)
{
#ifndef BEAGLE_NDEBUG
  if(mESVectorSize->size() == 0) {
    std::string lMessage = "GA::InitESVecOp::initIndividual: ";
    lMessage += "Size of the ES vector size parameter is zero. ";
    lMessage += "Could not initialize the individuals!";
    throw Beagle_RunTimeExceptionM(lMessage);
  }
  for(unsigned int k=0; k<mESVectorSize->size(); k++) {
    if((*mESVectorSize)[k] == 0) {
      std::string lMessage = "GA::InitESVecOp::initIndividual: Value of the ";
      lMessage += uint2ordinal(k);
      lMessage += " ES vector size parameter is zero. ";
      lMessage += "Could not initialize the individuals!";
      throw Beagle_RunTimeExceptionM(lMessage);
    }
  }
#endif // BEAGLE_NDEBUG

  Beagle_LogVerboseM(
    ioContext.getSystem().getLogger(),
    "initialization", "Beagle::GA::InitESVecOp",
    std::string("Resizing the individual to ")+
    uint2str(mESVectorSize->size())+" ES vectors"
  );
  outIndividual.resize(mESVectorSize->size());
  for(unsigned int i=0; i<outIndividual.size(); ++i) {
    GA::ESVector::Handle lESVector = castHandleT<GA::ESVector>(outIndividual[i]);
    lESVector->resize((*mESVectorSize)[i]);
    for(unsigned int j=0; j<lESVector->size(); ++j) {
      (*lESVector)[j].mValue =
        ioContext.getSystem().getRandomizer().rollGaussian(0.0,
                                                           mInitStrategyValue->getWrappedValue());
      (*lESVector)[j].mStrategy = mInitStrategyValue->getWrappedValue();                                                   
    }
    Beagle_LogVerboseM(
      ioContext.getSystem().getLogger(),
      "initialization", "Beagle::GA::InitESVecOp",
      std::string("ES vector ")+uint2str(i+1)+
      std::string(" has ")+uint2str((*mESVectorSize)[i])+" initialized values"
    );
    Beagle_LogDebugM(
      ioContext.getSystem().getLogger(),
      "initialization", "Beagle::GA::InitESVecOp",
      uint2ordinal(i+1)+
      std::string(" ES vector is initialized as: ")+lESVector->serialize()
    );
  }
}

