/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/GA/src/InitFltVecOp.cpp
 *  \brief  Source code of class GA::InitFltVecOp.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.8 $
 *  $Date: 2004/07/04 16:58:22 $
 */

#include "beagle/GA.hpp"

#include <sstream>

using namespace Beagle;


/*!
 *  \brief Construct a GA float vectors uniformly distributed initialization operator.
 *  \param inFloatVectorSize Size of the float vectors initialized.
 *  \param inReproProbaName Reproduction probability parameter name used in register. 
 *  \param inName Name of the operator.
 */
GA::InitFltVecOp::InitFltVecOp(IntegerVector inFloatVectorSize,
                               std::string inReproProbaName,
                               std::string inName) :
  InitializationOp(inReproProbaName, inName),
  mFloatVectorSize(new IntegerVector(inFloatVectorSize))
{ }


/*!
 *  \brief Initialize the GA float vectors uniformly distributed initialization operator.
 *  \param ioSystem System of the evolution.
 */
void GA::InitFltVecOp::initialize(System& ioSystem)
{
  InitializationOp::initialize(ioSystem);

  if(ioSystem.getRegister().isRegistered("ga.init.vectorsize")) {
    mFloatVectorSize = castHandleT<IntegerVector>(ioSystem.getRegister()["ga.init.vectorsize"]);
  } else {
    std::string lDefaultValue;
    for(unsigned int i=0; i<mFloatVectorSize->size(); i++) {
      lDefaultValue += uint2str((*mFloatVectorSize)[i]);
      if(i != (mFloatVectorSize->size()-1)) lDefaultValue += "/";
    }
    std::string lLongDescrip("Float vector size for individuals. The size of the ");
    lLongDescrip += "IntegerVector is the number of vectors composing the GA individuals, ";
    lLongDescrip += "while each value of the vector is the number of floats composing each vector. ";
    lLongDescrip += "IntegerVector format is S1/S2/.../Sn, where Si is ith value.";
    Register::Description lDescription(
      "Initial float vectors sizes",
      "IntegerVector",
      lDefaultValue,
      lLongDescrip
    );
    ioSystem.getRegister().addEntry("ga.init.vectorsize", mFloatVectorSize, lDescription);
  }

  if(ioSystem.getRegister().isRegistered("ga.init.maxvalue")) {
    mMaxInitValue = castHandleT<Float>(ioSystem.getRegister()["ga.init.maxvalue"]);
  } else {
    mMaxInitValue = new Float(1.0);
    Register::Description lDescription(
      "Max floats initialization value",
      "Float",
      "1.0",
      "Maximum initialization value assigned to vector's floats."
    );
    ioSystem.getRegister().addEntry("ga.init.maxvalue", mMaxInitValue, lDescription);
  }

  if(ioSystem.getRegister().isRegistered("ga.init.minvalue")) {
    mMinInitValue = castHandleT<Float>(ioSystem.getRegister()["ga.init.minvalue"]);
  } else {
    mMinInitValue = new Float(-1.0);
    Register::Description lDescription(
      "Min floats initialization value",
      "Float",
      "-1.0",
      "Minimum initialization value assigned to vector's floats."
    );
    ioSystem.getRegister().addEntry("ga.init.minvalue", mMinInitValue, lDescription);
  }
}


/*!
 *  \brief Initialize real-valued GA individual with numbers uniformly distributed in a given range.
 *  \param outIndividual Individual to initialize.
 *  \param ioContext Evolution context.
 */
void GA::InitFltVecOp::initIndividual(Beagle::Individual& outIndividual, Context& ioContext)
{
#ifndef BEAGLE_NDEBUG
  if(mFloatVectorSize->size() == 0) {
    std::string lMessage = "GA::InitFltVecOp::initIndividual: ";
    lMessage += "Size of the float vectors size parameter is zero. ";
    lMessage += "Could not initialize the individuals!";
    throw Beagle_RunTimeExceptionM(lMessage);
  }
  for(unsigned int k=0; k<mFloatVectorSize->size(); k++) {
    if((*mFloatVectorSize)[k] == 0) {
      std::string lMessage = "GA::InitFltVecOp::initIndividual: Value of the ";
      lMessage += uint2ordinal(k);
      lMessage += " float vector size parameter is zero. ";
      lMessage += "Could not initialize the individuals!";
      throw Beagle_RunTimeExceptionM(lMessage);
    }
  }
#endif // BEAGLE_NDEBUG

  Beagle_LogVerboseM(
    ioContext.getSystem().getLogger(),
    "initialization", "Beagle::GA::InitFltVecOp",
    std::string("Resizing the individual to ")+
    uint2str(mFloatVectorSize->size())+" float vectors"
  );
  outIndividual.resize(mFloatVectorSize->size());
  for(unsigned int i=0; i<outIndividual.size(); i++) {
    GA::FloatVector::Handle lFloatVector = castHandleT<GA::FloatVector>(outIndividual[i]);
    lFloatVector->resize((*mFloatVectorSize)[i]);
    for(unsigned int j=0; j<lFloatVector->size(); j++) {
      (*lFloatVector)[j] =
        ioContext.getSystem().getRandomizer().rollUniform(mMinInitValue->getWrappedValue(),
                                                          mMaxInitValue->getWrappedValue());
    }
    Beagle_LogVerboseM(
      ioContext.getSystem().getLogger(),
      "initialization", "Beagle::GA::InitFltVecOp",
      std::string("Float vector ")+uint2str(i+1)+
      std::string(" has ")+uint2str((*mFloatVectorSize)[i])+" initialized values"
    );
    Beagle_LogDebugM(
      ioContext.getSystem().getLogger(),
      "initialization", "Beagle::GA::InitFltVecOp",
      uint2ordinal(i+1)+
      std::string(" float vector is initialized as: ")+lFloatVector->serialize()
    );
  }
}

