/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/GA/src/MutationGaussianFltVecOp.cpp
 *  \brief  Source code of class GA::MutationGaussianFltVecOp.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.10 $
 *  $Date: 2004/04/25 23:12:19 $
 */

#include "beagle/GA.hpp"

#include <cfloat>
#include <float.h>    // To get FLT_MAX on Borland C++ Builder
#include <algorithm>
#include <string>

using namespace Beagle;


/*!
 *  \brief Construct a real-valued GA Gaussian mutation operator.
 *  \param inMutationPbName Gaussian mutation probability parameter name used in register.
 *  \param inMutateFloatPbName Single value Gaussian mutation probability name used in the register.
 *  \param inMutateGaussMuName Gaussian mutation mean value parameter name used in register.
 *  \param inMutateGaussSigmaName Gaussian mutation std deviation parameter name used in register.
 *  \param inName Name of the operator.
 */
GA::MutationGaussianFltVecOp::MutationGaussianFltVecOp(std::string inMutationPbName,
                                                       std::string inMutateFloatPbName,
                                                       std::string inMutateGaussMuName,
                                                       std::string inMutateGaussSigmaName,
                                                       std::string inName) :
  MutationOp(inMutationPbName, inName),
  mMutateFloatPbName(inMutateFloatPbName),
  mMutateGaussMuName(inMutateGaussMuName),
  mMutateGaussSigmaName(inMutateGaussSigmaName)
{ }


/*!
 *  \brief Initialize the real-valued GA Gaussian mutation operator.
 *  \param ioSystem System of the evolution.
 */
void GA::MutationGaussianFltVecOp::initialize(System& ioSystem)
{
  MutationOp::initialize(ioSystem);

  if(ioSystem.getRegister().isRegistered(mMutationPbName)) {
    ioSystem.getRegister().deleteEntry(mMutationPbName);
  }
  
  if(ioSystem.getRegister().isRegistered(mMutationPbName)) {
    mMutationProba = castHandleT<Float>(ioSystem.getRegister()[mMutationPbName]);
  } else {
    mMutationProba = new Float(float(1.0));
    Register::Description lDescription(
      "Indiv. Gaussian mutation prob.",
      "Float",
      "1.0",
      "Gaussian mutation probability for each real-valued GA individual."
    );
    ioSystem.getRegister().addEntry(mMutationPbName, mMutationProba, lDescription);
  }

  if(ioSystem.getRegister().isRegistered(mMutateFloatPbName)) {
    mMutateFloatPb = castHandleT<Float>(ioSystem.getRegister()[mMutateFloatPbName]);
  } else {
    mMutateFloatPb = new Float(float(0.1));
    Register::Description lDescription(
      "Value's Gaussian mutation prob.",
      "Float",
      "0.1",
      "Single value Gaussian mutation probability for the real-valued GA individual."
    );
    ioSystem.getRegister().addEntry(mMutateFloatPbName, mMutateFloatPb, lDescription);
  }
  
  if(ioSystem.getRegister().isRegistered(mMutateGaussMuName)) {
    mMutateGaussMu = castHandleT<Float>(ioSystem.getRegister()[mMutateGaussMuName]);
  } else {
    mMutateGaussMu = new Float(float(0.0));
    Register::Description lDescription(
      "Gaussian mutation mean",
      "Float",
      "0.0",
      "Mean of values added to real-values vector numbers at each Gaussian mutation."
    );
    ioSystem.getRegister().addEntry(mMutateGaussMuName, mMutateGaussMu, lDescription);
  }

  if(ioSystem.getRegister().isRegistered(mMutateGaussSigmaName)) {
    mMutateGaussSigma = castHandleT<Float>(ioSystem.getRegister()[mMutateGaussSigmaName]);
  } else {
    mMutateGaussSigma = new Float(float(0.1));
    Register::Description lDescription(
      "Gaussian mutation std deviation",
      "Float",
      "0.1",
      "Standard deviation of values added to real-values vector numbers at each Gaussian mutation."
    );
    ioSystem.getRegister().addEntry(mMutateGaussSigmaName, mMutateGaussSigma, lDescription);
  }

  if(ioSystem.getRegister().isRegistered("ga.float.maxvalue")) {
    mFloatsValueMax = castHandleT<Float>(ioSystem.getRegister()["ga.float.maxvalue"]);
  } else {
    mFloatsValueMax = new Float(float(FLT_MAX));
    Register::Description lDescription(
      "Maximum number value in vectors",
      "Float",
      dbl2str(FLT_MAX),
      "Maximum value of real numbers in the genotype float vectors."
    );
    ioSystem.getRegister().addEntry("ga.float.maxvalue", mFloatsValueMax, lDescription);
  }

  if(ioSystem.getRegister().isRegistered("ga.float.minvalue")) {
    mFloatsValueMin = castHandleT<Float>(ioSystem.getRegister()["ga.float.minvalue"]);
  } else {
    mFloatsValueMin = new Float(float(FLT_MIN));
    Register::Description lDescription(
      "Minimum number value in vectors",
      "Float",
      dbl2str(FLT_MIN),
      "Minimum value of real numbers in the genotype float vectors."
    );
    ioSystem.getRegister().addEntry("ga.float.minvalue", mFloatsValueMin, lDescription);
  }
}


/*!
 *  \brief Gaussian mutate a real-valued GA individual.
 *  \param ioIndividual Real-valued GA individual to mutate.
 *  \param ioContext Context of the evolution.
 *  \return True if the individual is effectively mutated, false if not.
 */
bool GA::MutationGaussianFltVecOp::mutate(Beagle::Individual& ioIndividual, Context& ioContext)
{
  Beagle_ValidateParameterM(mMutateGaussSigma->getWrappedValue()>=0.0,mMutateGaussSigmaName,"<0");
  
  bool lMutated = false;
  Beagle_LogVerboseM(
    ioContext.getSystem().getLogger(),
    "mutation", "Beagle::GA::MutationGaussianFltVecOp",
    std::string("Gaussian mutations with mean of ")+
    dbl2str(mMutateGaussMu->getWrappedValue())+
    std::string(", and standard deviation of ")+
    dbl2str(mMutateGaussSigma->getWrappedValue())
  );
  
  for(unsigned int i=0; i<ioIndividual.size(); i++) {
    GA::FloatVector::Handle lVector = castHandleT<GA::FloatVector>(ioIndividual[i]);
    Beagle_LogVerboseM(
      ioContext.getSystem().getLogger(),
      "mutation", "Beagle::GA::MutationGaussianFltVecOp",
      std::string("Gaussian mutation the ")+uint2ordinal(i+1)+" float vector"
    );
    Beagle_LogDebugM(
      ioContext.getSystem().getLogger(),
      "mutation", "Beagle::GA::MutationGaussianFltVecOp",
      std::string("The float vector before GA Gaussian mutation: ")+lVector->serialize()
    );   
    for(unsigned int j=0; j<lVector->size(); j++) {
      double lRolledPb = ioContext.getSystem().getRandomizer().rollUniform();
      if(lRolledPb <= mMutateFloatPb->getWrappedValue()) {
        float lMValue = ioContext.getSystem().getRandomizer().rollGaussian(
                          mMutateGaussMu->getWrappedValue(), mMutateGaussSigma->getWrappedValue());
        (*lVector)[j] += lMValue;
        if((*lVector)[j] > mFloatsValueMax->getWrappedValue())
          (*lVector)[j] = mFloatsValueMax->getWrappedValue();
        if((*lVector)[j] < mFloatsValueMin->getWrappedValue())
          (*lVector)[j] = mFloatsValueMin->getWrappedValue();
        lMutated = true;
        Beagle_LogDebugM(
          ioContext.getSystem().getLogger(),
          "mutation", "Beagle::GA::MutationGaussianFltVecOp",
          std::string("Gaussian mutating by adding ")+dbl2str(lMValue)+
          std::string(" to the value at the index ")+uint2str(j)+
          std::string(" of the float vector")
        );
      }
    }
    if(lMutated) {
      Beagle_LogVerboseM(
        ioContext.getSystem().getLogger(),
        "mutation", "Beagle::GA::MutationGaussianFltVecOp",
        std::string("The float vector has been mutated")
      );
      Beagle_LogDebugM(
        ioContext.getSystem().getLogger(),
        "mutation", "Beagle::GA::MutationGaussianFltVecOp",
        std::string("The float vector after GA Gaussian mutation: ")+lVector->serialize()
      );
    }
    else {
      Beagle_LogVerboseM(
        ioContext.getSystem().getLogger(),
        "mutation", "Beagle::GA::MutationGaussianFltVecOp",
        std::string("The float vector has not been mutated")
      );
    }
  }
  return lMutated;
}


/*!
 *  \brief Read a Gaussian mutation operator from XML tree.
 *  \param inNode XML node to use to read mutation operator.
 *  \param inOpMap Operator map to use to read mutation operator.
 */
void GA::MutationGaussianFltVecOp::readWithMap(XMLNode::Handle& inNode, OperatorMap& inOpMap)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != getName())) {
    std::ostringstream lOSS;
    lOSS << "tag <" << getName() << "> expected!" << std::flush;
    throw Beagle_IOExceptionNodeM(*inNode, lOSS.str());
  }
  std::string lMutationPbReadName = inNode->getTagAttribute("mutationpb");
  if(lMutationPbReadName.empty() == false) mMutationPbName = lMutationPbReadName;
  std::string lMutateFloatPbReadName = inNode->getTagAttribute("mutfloatpb");
  if(lMutateFloatPbReadName.empty() == false) mMutateFloatPbName = lMutateFloatPbReadName;
  std::string lMutateGaussMuReadName = inNode->getTagAttribute("mutgaussmu");
  if(lMutateGaussMuReadName.empty() == false) mMutateGaussMuName = lMutateGaussMuReadName;
  std::string lMutateGaussSigmaReadName = inNode->getTagAttribute("mutgausssigma");
  if(lMutateGaussSigmaReadName.empty() == false) mMutateGaussSigmaName = lMutateGaussSigmaReadName;  
}


/*!
 *  \brief Write Gaussian mutation operator into XML streamer.
 *  \param ioStreamer XML streamer to write mutation operator into.
 */
void GA::MutationGaussianFltVecOp::writeContent(XMLStreamer& ioStreamer) const
{
  Beagle::MutationOp::writeContent(ioStreamer);
  ioStreamer.insertAttribute("mutfloatpb", mMutateFloatPbName);
  ioStreamer.insertAttribute("mutgaussmu", mMutateGaussMuName);
  ioStreamer.insertAttribute("mutgausssigma", mMutateGaussSigmaName);
}


