/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/RandomizerNR.cpp
 *  \brief  Source code of class RandomizerNR.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.4 $
 *  $Date: 2004/04/25 23:12:21 $
 */

#include "beagle/Beagle.hpp"

#include <ctime>

using namespace Beagle;


/*!
 *  \brief Construct a randomizer NR.
 */
RandomizerNR::RandomizerNR() :
  mIY(0),
  mIDUM(0)
{ }


/*!
 *  \brief Initialize the randomizer state.
 *  \param inState Value of the state. If 0, set to seed value (if seed 0, set to timer value).
 */
void RandomizerNR::initRandomizer(unsigned long inState)
{
  Beagle_AssertM(mSeed->getWrappedValue() >= 0);

  if(inState == 0) {
    if(mSeed->getWrappedValue() == 0) mIDUM = *mSeed = std::time(0);
    else mIDUM = *mSeed;
  }
  else mIDUM = inState;

  for(int j=Beagle_NTAB+7; j>=0; j--) {
    unsigned long lTemp = mIDUM / Beagle_IQ;
    mIDUM = Beagle_IA*(mIDUM - lTemp*Beagle_IQ) - Beagle_IR*lTemp;
    if(mIDUM < 0) mIDUM += Beagle_IM;
    if(j < Beagle_NTAB) mIV[j] = mIDUM;
  }
  mIY = mIV[0];
  *mState = mIDUM;
}


/*!
 *  \brief Read the state of the randomizer from a XML node.
 *  \param inNode XML node to read the randomizer NR from.
 */
void RandomizerNR::read(XMLNode::Handle& inNode)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != "Randomizer"))
    throw Beagle_IOExceptionNodeM(*inNode, "tag <Randomizer> expected!");

  std::string lTypeStr = inNode->getTagAttribute("type");
  if(lTypeStr != "NR")
    throw Beagle_IOExceptionNodeM(*inNode,"type of randomizer is not NR!");  

  for(XMLNode::Handle lChild=inNode->getFirstChild();
      lChild!=NULL; lChild=lChild->getNextSibling()) {
    if(lChild->getType() == XMLNode::eTag) {
      if(lChild->getTagName() == "Seed") {
        XMLNode::Handle lChild2 = lChild->getFirstChild();
        if(lChild2 == NULL)
          throw Beagle_IOExceptionNodeM(*lChild,"missing value of randomizer seed!");
        if(lChild2->getType() != XMLNode::eString)
          throw Beagle_IOExceptionNodeM(*lChild2,"missing value of randomizer seed!");
        *mSeed = str2int(lChild2->getValue());
      }
      else if(lChild->getTagName() == "State") {
        XMLNode::Handle lChild2 = lChild->getFirstChild();
        if(lChild2 == NULL)
          throw Beagle_IOExceptionNodeM(*lChild,"missing value of randomizer state!");
        if(lChild2->getType() != XMLNode::eString)
          throw Beagle_IOExceptionNodeM(*lChild2,"missing value of randomizer state!");
        *mState = str2int(lChild2->getValue());
      }
      else if(lChild->getTagName() == "IV") {
        XMLNode::Handle lChild2 = lChild->getFirstChild();
        if(lChild2 == NULL) throw Beagle_IOExceptionNodeM(*lChild,"missing IV array!");
        if(lChild2->getType() != XMLNode::eString)
          throw Beagle_IOExceptionNodeM(*lChild2,"missing IV array!");
        std::istringstream lISS(lChild2->getValue());
        char lTmp = '\0';
        for(int i=0; i<(Beagle_NTAB-1); ++i) {
          lISS >> mIV[i] >> lTmp;
          if(lTmp != '/')
            throw Beagle_IOExceptionNodeM(*lChild2, "bad delimiter in IV enumeration!");
        }
        lISS >> mIV[Beagle_NTAB-1];
      }
      else if(lChild->getTagName() == "IY") {
        XMLNode::Handle lChild2 = lChild->getFirstChild();
        if(lChild2 == NULL) throw Beagle_IOExceptionNodeM(*lChild,"missing IY value!");
        if(lChild2->getType() != XMLNode::eString)
          throw Beagle_IOExceptionNodeM(*lChild2,"missing IY value!");
        mIY = str2int(lChild2->getValue());
      }
      else if(lChild->getTagName() == "IDUM") {
        XMLNode::Handle lChild2 = lChild->getFirstChild();
        if(lChild2 == NULL) throw Beagle_IOExceptionNodeM(*lChild,"missing IDUM value!");
        if(lChild2->getType() != XMLNode::eString)
          throw Beagle_IOExceptionNodeM(*lChild2,"missing IDUM value!");
        mIDUM = str2int(lChild2->getValue());
      }
    }
  }
}


/*!
 *  \brief Generate a integer following an uniform discrete distribution.
 *  \param inLower Lower bound of the distribution.
 *  \param inUpper Upper bound of the distribution.
 *  \return Random number generated.
 */
unsigned long RandomizerNR::rollInteger(unsigned long inLower, unsigned long inUpper)
{
  Beagle_AssertM(inLower <= inUpper);
  unsigned long lInteger = (unsigned long)(inLower + (inUpper - inLower + 1) * rollUniform());
  if(lInteger > inUpper) lInteger = inUpper;
  return lInteger;
}


/*!
 *  \brief Generate a number following a uniform distribution in the interval [inLower,inUpper).
 *  \param inLower Lower bound of the uniform distribution.
 *  \param inUpper Upper bound of the uniform distribution.
 *  \return Random number generated in [inLower,inUpper).
 */
double RandomizerNR::rollUniform(double inLower, double inUpper)
{
  Beagle_AssertM(inUpper > inLower);
  if((mState->getWrappedValue()==0) || (mState->getWrappedValue()!=mIDUM)) initRandomizer(*mState);
  double lTemp;
  unsigned long lK = mIDUM / Beagle_IQ;
  mIDUM = Beagle_IA*(mIDUM - lK*Beagle_IQ) - Beagle_IR*lK;
  if(mIDUM < 0) mIDUM += Beagle_IM;
  int lJ = (int)(mIY/Beagle_NDIV);
  mIY = mIV[lJ];
  *mState = mIV[lJ] = mIDUM;
  if((lTemp = Beagle_AM*mIY) > Beagle_RNMX) return (Beagle_RNMX*(inUpper-inLower) + inLower);
  else return (lTemp*(inUpper-inLower) + inLower);
}


/*!
 *  \brief Write the state of the randomizer into a Beagle XML streamer.
 *  \param ioStreamer XML streamer to write the randomizer into.
 */
void RandomizerNR::write(XMLStreamer& ioStreamer) const
{
  ioStreamer.openTag("Randomizer");
  ioStreamer.insertAttribute("type", "NR");
  ioStreamer.insertInteger(mSeed->getWrappedValue(), "Seed");
  ioStreamer.insertInteger(mState->getWrappedValue(), "State");
  ioStreamer.insertInteger(mIY, "IY");
  ioStreamer.insertInteger(mIDUM, "IDUM");
  std::ostringstream lOSS;
  for(unsigned int i=0; i<Beagle_NTAB-1; i++) lOSS << mIV[i] << "/";
  lOSS << mIV[Beagle_NTAB-1];
  ioStreamer.insertString(lOSS.str(), "IV");
  ioStreamer.closeTag();
}


